import calendar
import json
import math
from datetime import datetime
from typing import List
from typing import Optional

from typeguard import typechecked

from repository.AbstractRepository import AbstractRepository
from model.Sensor import Sensor
from model.SensorDataPoint import SensorDataPoint
from model.SensorProperties import SensorProperties
from tools.RetryPattern import RetryPattern


class SensorsRepository(AbstractRepository):
    __REDIS_SENSORS_KEY = 'sensors'
    __REDIS_SENSORS_HISTORY_KEY = 'sensors_history_{0}'
    __SENSORS_UPDATE_INTERVAL_IN_HISTORY = 300

    @typechecked()
    def __init__(self, redis_configuration: dict, retry_pattern: RetryPattern):
        AbstractRepository.__init__(self, redis_configuration)
        self.__retry_pattern = retry_pattern
        self.keys = {self.__REDIS_SENSORS_KEY: {}}
        self.last_averages = {}
        self.sensors_last_updated = {}
        self.current_timestamp = 0

    @typechecked()
    def get_sensors(self) -> List[Sensor]:
        sensors_data = self.get(self.__REDIS_SENSORS_KEY)
        sensors = []
        for sensor_data in sensors_data:
            # todo remove this hack after all sensors are updated
            last_updated = sensor_data['last_updated'] if 'last_updated' in sensor_data else 0
            sensor = Sensor(sensor_data['id'], sensor_data['type'], sensor_data['location'],
                            sensor_data['value'], sensor_data['device_type'], last_updated)
            if 'properties' in sensor_data:
                sensor.properties = self.__get_sensor_properties(sensor_data['properties'])
            else:
                sensor.properties = self.__get_sensor_properties({})
            sensors.append(sensor)

        return sensors

    @typechecked()
    def get_sensor(self, id: str) -> Optional[Sensor]:
        sensors = self.get_sensors()
        sensor = [sensor for sensor in sensors if sensor.id == id]
        if len(sensor) == 1:
            return sensor[0]

        return None

    @typechecked()
    def set_sensor(self, sensor: Sensor) -> None:
        # @todo this current_timestamp is fishy, please investigate
        self.current_timestamp = calendar.timegm(datetime.now().timetuple())
        self.__set(sensor)
        self.__add_last_sensor_averages_in_history(sensor)

    @typechecked()
    def set_sensors(self, sensors: list):
        self.client.set(self.__REDIS_SENSORS_KEY, json.dumps(sensors))

    @typechecked()
    def get_sensor_values(self, id: str, start_date: datetime, end_date: datetime) -> List[SensorDataPoint]:
        redis_key = self.__REDIS_SENSORS_HISTORY_KEY.format(id)
        start_timestamp = calendar.timegm(start_date.timetuple())
        end_timestamp = calendar.timegm(end_date.timetuple())
        data = self.client.zrangebyscore(redis_key, start_timestamp, end_timestamp, withscores=True)

        return [SensorDataPoint(json.loads(datapoint[0].decode("utf-8"))['value'], datapoint[1]) for datapoint in data]

    def __set(self, sensor: Sensor):
        sensors = self.get(self.__REDIS_SENSORS_KEY)
        for redis_sensor in sensors:
            if redis_sensor['id'] == sensor.id:
                redis_sensor['value'] = sensor.value
                redis_sensor['last_updated'] = sensor.last_updated
        def wrapper():
            self.client.set(self.__REDIS_SENSORS_KEY, json.dumps(sensors))
        self.__retry_pattern.run(wrapper, 3)

    def __add_last_sensor_averages_in_history(self, sensor: Sensor):
        name = self.__get_sensor_key(sensor.id)
        if name not in list(self.last_averages.keys()):
            self.last_averages[name] = []
        self.last_averages[name].append(sensor.value)
        if name not in self.sensors_last_updated:
            self.sensors_last_updated[name] = 0
        if self.current_timestamp - self.sensors_last_updated[name] < self.__SENSORS_UPDATE_INTERVAL_IN_HISTORY:
            return

        self.sensors_last_updated[name] = self.current_timestamp
        sensor_average_value = int(math.ceil(float(sum(self.last_averages[name])) / len(self.last_averages[name])))
        self.add_to_list(self.__get_sensor_key(sensor.id), {'value': sensor_average_value})
        self.last_averages[name] = []

    def __get_sensor_key(self, id: str) -> str:
        return self.__REDIS_SENSORS_HISTORY_KEY.format(id)

    def __get_sensor_properties(self, properties: dict) -> SensorProperties:
        actuator_properties = SensorProperties()
        for name, value in properties.items():
            actuator_properties.set(name, value)

        return actuator_properties